# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
import string
import subprocess
import select
import logging


class dsync_error_proto(Exception):
    """
    """
    pass


# Line terminators (we always output CRLF, but accept any of CRLF, LFCR, LF)
CR = '\r'
LF = '\n'
CRLF = CR+LF


class DsyncProtocol(object):
    '''
    classdocs
    '''


    def __init__(self):
        '''
        Constructor
        '''
        self.connection = None
        self.timeout = None
        self._debugging = 0
        self.logger = logging.getLogger("DsyncProtocol")
    
    def open(self, mailbox):
        pass
    
    def _settimeout(self, timeout):
        self.timeout = timeout
        
    def _write(self, data):
        self.stdin.write(data)
    
    def _getline(self):
        if (self.timeout):
            inputs = [self.stdin]
            outputs = []
            readable, writable, exceptional = select.select(inputs, outputs, inputs, self.timeout)
            if (not (readable or writable or exceptional) or (readable[0] != self.stdin)):
                raise dsync_error_proto('-ERR EOF')
        line = self.stdout.readline()
        if self._debugging > 1: print '*get*', repr(line)
        if not line: raise dsync_error_proto('-ERR EOF')
        octets = len(line)
        if line[-2:] == CRLF:
            if self._debugging > 1: print 'CRLF'
            return line[:-2], octets
        if line[0] == CR:
            if self._debugging > 1: print 'CR'
            return line[1:-1], octets
        if self._debugging > 1: print 'OTHER'
        return line[:-1], octets

    def _getresp(self):
        resp, o = self._getline()
        if self._debugging > 1: print '*resp*', repr(resp)
        c = resp[:1]
        return resp
    
    def _geterr(self):
        inputs = [self.stderr]
        outputs = []
        timeout = 0.0
        readable, writable, exceptional = select.select(inputs, outputs, inputs, timeout)
        if (readable or writable or exceptional):
            err_lines = []
            while ((readable or writable or exceptional) and (readable[0] == self.stderr)):
                line = self.stderr.readline()
                if not line: break
                err_lines.append(line)
                readable, writable, exceptional = select.select(inputs, outputs, inputs, 0.2)
            return err_lines
        else:
            return

    def _geterrwait(self):
        line = self.stderr.readline()
        if line[-2:] == CRLF:
            if self._debugging > 1: print 'CRLF'
            return line[:-2]
        if line[0] == CR:
            if self._debugging > 1: print 'CR'
            return line[1:-1]
        if self._debugging > 1: print 'OTHER'
        return line[:-1]

    def _getlongresp(self):
        list = []; octets = 0
        line, o = self._getline()
        if line and line[0]=='0':
            # failure
            return line, o
        while line != '.':
            if line[:2] == '..':
                o = o-1
                line = line[1:]
            octets = octets + o
            list.append(line)
            line, o = self._getline()
        return list, octets

    def _getlongresp0term(self):
        list = []; octets = 0
        line, o = self._getline()
        while line != '+':
            octets = octets + o
            list.append(line)
            line, o = self._getline()
        return list, octets

    def _putline(self, line):
        if self._debugging > 1: print '*put*', repr(line)
        self._write('%s%s' % (line, CRLF))

    def _putcmd(self, line):
        if self._debugging: print '*cmd*', repr(line)
        self._putline(line)

    def _shortcmd(self, line):
        self._putcmd(line)
        return self._getresp()

    def _longcmd(self, line):
        self._putcmd(line)
#        time.sleep(1)
        err = self._geterr()
        if err: return err
        return self._getlongresp()

    def _longcmd0term(self, line):
        self._putcmd(line)
        return self._getlongresp0term()

    def clearinfo(self):
        """return info from stderr if buffer filled

        """
        return self._geterr()
        
    def getinfoline(self):
        """return info from stderr waiting

        """
        return self._geterrwait()
        
    def hasinfo(self):
        inputs = [self.stderr]
        outputs = []
        timeout = 0.5
        readable, writable, exceptional = select.select(inputs, outputs, inputs, timeout)
        if (readable or writable or exceptional):
            return True
        else:
            return False
    
    def getinfo(self):
        """return parsed info objects from stderr waiting

        """
        infoline = self.getinfoline()
        self.logger.debug("getinfo: " + infoline)
        # parse info like
        # dsync-remote(ryan): Info: save: box=Drafts, uid=287, msgid=, size=29, flags=()
        dsync_remote, level, rest = infoline.split(':', 2)
        action = None
        paramsmap = None
        infovalues = None
        errortext = None
        level = level.strip()
        if level == 'Error':
            errortext = rest
        elif level == 'Warning':
            errortext = rest
        else:
            action, paramsstr = rest.split(':', 1)
            action = action.strip()
            params = paramsstr.split(',')
            if (len(params) > 1):
                paramsmap = {}
                for param in params:
                    key, value = param.split('=', 1)
                    paramsmap[key.strip()] = value
            else:
                infostr = params[0]
                if (infostr.find(' -> ') > -1):
                    infovalues = infostr.split(' -> ')
        return (dsync_remote, level, action, paramsmap, infovalues, errortext)
        
    def boxlist(self):
        """Request listing of foldes/boxes, return result.

        Result is in form
        ['box-info', ...], octets].
        """
        return self._longcmd0term('BOX-LIST')

    def msglist(self, boxguids):
        """Request listing of messages, return result.

        Result is in form
        ['msg-info', ...], octets].
        """
        return self._longcmd0term('MSG-LIST\t%s' % string.join(boxguids, '\t'))

    def msgget(self, boxguid, msguid):
        """Request message, return result.

        Result is in form
        ['message lines', ...], octets].
        """
        return self._longcmd('MSG-GET\t%s\t%s' % (boxguid, msguid))
        
    def boxselect(self, boxguid):
        """Select box/folder

        No Result.
        """
        self._putline('BOX-SELECT\t%s' % (boxguid))
        return self._geterr()
        
    def boxcreate(self, boxname, name_sep, last_change, flags, boxguid, uid_validity, uid_next, message_count, highest_modseq, first_recent_uid, cache_fields):
        """Create box/folder

        No Result.
        """
        self._putline('BOX-CREATE\t%s\t%s\t%d\t%d\t%s\t%d\t%d\t%d\t%d\t%d\t%s' % (
            boxname, name_sep, last_change, flags, boxguid, uid_validity, uid_next, message_count, highest_modseq, first_recent_uid, cache_fields))
        return self._geterr()
        
    def boxrename(self, boxguid, boxname, name_sep):
        """Rename box/folder

        No Result.
        """
        self._putline('BOX-RENAME\t%s\t%s\t%s' % (
            boxguid, boxname, name_sep))
        return self._geterr()
        
    def boxdelete(self, boxguid, last_change):
        """Delete box/folder

        No Result.
        """
        self._putline('BOX-DELETE\t%s\t%d' % (
            boxguid, last_change)) 
        return self._geterr()
        
    def msgsave(self, received_date, pop3uidl, msgguid, msguid, modseq, flags, save_date, msglines):
        """Save message

        No Result.
        """
        self._putline('MSG-SAVE\t%s\t%s\t%s\t%s\t%s\t%s\t%s' % 
                             (received_date, pop3uidl, msgguid, msguid, modseq, flags, save_date))
        err = self._geterr()
        if err: return err
        for msgline in msglines:
            if msgline[:1] == '.':
                self._putline('.%s' % msgline)
            else:
                self._putline(msgline)
            err = self._geterr()
            if err: return err
        self._putline('.')
        err = self._geterr()
        if err: return err

    def msgupdate(self, uid, modseq, flags):
        """Create box/folder

        No Result.
        """
        self._putline('MSG-UPDATE\t%d\t%d\t%s' % (uid, modseq, flags))  
        return self._geterr()
        
    def msgexpunge(self, uid):
        """Create box/folder

        No Result.
        """
        self._putline('MSG-EXPUNGE\t%d' % (uid))  
        return self._geterr()
        
    def finish(self):
        """Request finish, return result.

        Result is in form
        ['ok', ...], octets].
        """
        try:
            return self._shortcmd('FINISH')
        except:
            return(self._geterr())
            return('-ERR EOF')

    def connect(self):
        pass

    def open(self, mailbox):
        if (self.connection != None):
            if self._debugging: self.logger.debug("dsync still open!")
            self.close()
            
        bufsize = 1
        client = subprocess.Popen(["dsync", "-v", "-u", mailbox, "server"], bufsize=bufsize,
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        self.stdin = client.stdin
        self.stdout = client.stdout
        self.stderr = client.stderr
        self.connection = client
        try:
            server_version = self._getline()
            if self._debugging: self.logger.debug('*server_version* ' + repr(server_version))
            # TODO: check server version
            self._putline("dsync-client\t1")
        except:
            if self._debugging: self.logger.debug('open dsync error:' + self._geterr())
            returncode = self.connection.wait()
            if self._debugging: self.logger.debug('open dsync returncode:' + repr(returncode))
            raise dsync_error_proto('-ERR EOF')

    def close(self):
        returncode = None
        if (self.connection != None):
            if (self.stdin != None):
                self.stdin.close()
                returncode = self.connection.wait()
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self.connection = None
        
        if self._debugging: self.logger.debug("RETURNCODE: " + repr(returncode))
        return returncode

    def close_channel(self):
        self.close()

    def reconnect(self):
        self.close()
        self.connect()

    def reopen(self, mailbox):
        returncode = self.close()
        if (returncode):
            return returncode
        try:
            self.open(mailbox)
        except:
            self.open(mailbox)

        

        
try:
    from paramiko import SSHClient
except ImportError:
    pass
else:
    class DsyncProtocolSSH(DsyncProtocol):
        '''
        classdocs
        '''
    
    
        def __init__(self, hostname=None, password=None, username=None):
            '''
            Constructor
            '''
            super(DsyncProtocolSSH, self).__init__()

            self.hostname = hostname
            self.username = username
            self.password = password
            self.channel = None

        
        def connect(self):
            # TODO: handle unknown hosts    
            client = SSHClient()
            client.load_system_host_keys()
            client.connect(hostname=self.hostname,username=self.username,password=self.password)
            self.connection = client

        def close(self):
            if (self.channel):
                self.channel.close()
                self.channel = None
            if (self.connection):
                self.connection.close()
                self.connection = None
                
        def reconnect(self):
            self.close()
            self.connect()
            
        def open(self, mailbox):
            bufsize = -1
            transport = self.connection.get_transport()
            chan = transport.open_session()
            self.channel = chan
            chan.exec_command("dsync -v -u %s server" % mailbox)
            self.stdin = chan.makefile('wb', bufsize)
            self.stdout = chan.makefile('rb', bufsize)
            self.stderr = chan.makefile_stderr('rb', bufsize)
            server_version = self._getline()
            if self._debugging: print '*server_version*', repr(server_version)
            # TODO: check server version
            self._putline("dsync-client\t1")

        def close_channel(self):
            if (self.channel):
                self.channel.close()
                self.channel.recv_exit_status()
                self.channel = None
                # TODO: close stdin, stdout and stderr
                
        def reopen(self, mailbox):
            try:
                self.close_channel()
                self.open(mailbox)
            except:
                self.connect()
                self.open(mailbox)
            
        def _geterr(self):
            if self.channel.recv_stderr_ready():
                err_lines = []
                while self.channel.recv_stderr_ready():
                    err_lines.append(self.stderr.readline())
                return err_lines
            else:
                return

        def hasinfo(self):
            return self.channel.recv_stderr_ready()

        def _settimeout(self, timeout):
            self.channel.settimeout(timeout)
    
            
