# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from mailbridge.store.BaseStore import BaseStore
from mailbridge.store.dsync.DsyncMailbox import DsyncMailbox
from mailbridge.store.dsync.DsyncImporters import DsyncContentsImporter
from mailbridge.store.dsync.DsyncExporters import DsyncExportChangesContents
from mailbridge.store.dsync.DsyncExporters import DsyncExportChangesHierarchy
from mailbridge.protocol.DsyncProtocol import DsyncProtocol, DsyncProtocolSSH
from mailbridge.store.zarafa.ZarafaStoreExporters import ZarafaStoreExporterProxyContentsImporter, ZarafaStoreExporterProxyHierarchyImporter

import os
import redis

class DsyncStore(BaseStore):
    '''
    classdocs
    '''


    def __init__(self, settings_dict):
        '''
        Constructor
        '''
        super(DsyncStore, self).__init__(settings_dict)
        self.state = None
        self.mailbox_state = {}
        self.lastuid=0
        self.DraftsFolderId = None
        self.ReceiveFolderId = None
        self.WastebasketFolderId = None
        self.SentFolderId = None
        self.NewReceiveFolderId = None
        self.replica_id = None
        self.redis = None
        self.guid2pathname_map = {}
        self.pathname2guid_map = {}
        self.host=None
        self.version = 1
        
    def connect(self, hostname, username, password, account):
        connection = DsyncProtocol()
        connection.username = username
        connection.password = password
        connection.connect()
        connection.open(account)
        self.connection = connection
        self.account = account
        self._getStoreStateData()
        if (self.redis == None):
            self.redis = redis.Redis(host='localhost', port=6379, db=0)
        self._get_replica_id()

    def close_connection(self):
        self.connection.close()

    def reconnect(self):
        self.connection.reconnect()
        self.connection.open(self.account)
        self._getStoreStateData()
        self.lastuid=0
        
    def reopen(self):
        self.connection.reopen(self.account)
        self._getStoreStateData()
        self.lastuid=0

    def get_redis_basekey(self, replica_id=None):
        if replica_id==None:
            replica_id = self.replica_id
        if self.host==None:
            host = ""
        return "%s:%s:%s:%s" % (self.version, host, self.account, replica_id)
    
    def _get_replica_id(self):
        if (self.replica_id != None):
            return self.replica_id
        inbox_guid = self.getReceiveFolderId()
        if (inbox_guid == None):
            next_replica_id = self.redis.incr(self.get_redis_basekey(0)+":last_replica_id")
            self.replica_id = next_replica_id
        else:
            replica_id = self.redis.hget(self.get_redis_basekey(0)+":replica_id", inbox_guid)
            if replica_id != None:
                self.replica_id = int(replica_id)
            else:
                next_replica_id = self.redis.incr(self.get_redis_basekey(0)+":last_replica_id")
                self.replica_id = next_replica_id
    
    def _set_replica_id(self, inbox_guid, replica_id=None):
        if replica_id == None:
            replica_id = self.replica_id
        self.redis.hset(self.get_redis_basekey(0)+":replica_id", inbox_guid, replica_id)

    def getHierarchyImporter(self):
        importer = ZarafaStoreExporterProxyHierarchyImporter(self)
        return importer
    
    def getContentsImporter(self, id):
        folder = self.getFolder(id)
        proxy = ZarafaStoreExporterProxyContentsImporter()
        proxy.importer = DsyncContentsImporter(folder)
        drafts_folder_id = self.getDraftsFolderId()
        if id == drafts_folder_id:
            proxy.is_drafts_folder = 1
        return proxy
    
    def getExporter(self, id=None):
        if (id != None):
            folder = self.getFolder(id)
            exporter = DsyncExportChangesContents(folder)
        else:
            exporter = DsyncExportChangesHierarchy(self)
        return exporter
    
    def getFolderIds(self):
        ids = self.box_list_map.keys()
        return ids
    
    def getFolder(self, id):
        return DsyncMailbox(self, id)
        
    def getFolderPath(self, id):
        if (self.box_list_map.has_key(id)):
            return self.box_list_map[id][0]
        elif (self.guid2pathname_map.has_key(id)):
            return self.guid2pathname_map[id]
        else:
            return None
        
    def getFolderPaths(self):
        paths = self.box_list_path_map.keys()
        newpaths = self.pathname2guid_map.keys()
        for newpath in newpaths:
            try:
                paths.index(newpath)
            except:
                paths.append(newpath)
        return paths
    
    def getGUIDForFolderPath(self, pathname):
        if (self.box_list_path_map.has_key(pathname)):
            if (len(self.box_list_path_map[pathname]) > 4):
                return self.box_list_path_map[pathname][4]
            else:
                return None
        elif (self.pathname2guid_map.has_key(pathname)):
            return self.pathname2guid_map[pathname]
        else:
            return None
        
    def getFolderLastChange(self, id):
        if (self.box_list_map.has_key(id)):
            return self.box_list_map[id][2]
        
    def setFolderPath(self, id, pathname):
        self.guid2pathname_map[id] = pathname
        self.pathname2guid_map[pathname] = id
        
    def setFolderPathMapping(self, id, pathname, source_key):
        self.redis.hset(self.get_redis_basekey()+":guid2sourceid", id, source_key)
        self.redis.hset(self.get_redis_basekey()+":sourceid2guid", source_key, id)
        self.redis.hset(self.get_redis_basekey()+":guid2path", id, pathname)
        self.setFolderPath(id, pathname)

    def getGUIDs(self):
        return self.redis.hkeys(self.get_redis_basekey()+":guid2sourceid")
                        
    def getGUIDForSourceKey(self, source_key):
        return self.redis.hget(self.get_redis_basekey()+":sourceid2guid", source_key)
                        
    def get_folder_max_modseq(self, id):
        if (self.box_list_map.has_key(id)):
            return int(self.box_list_map[id][8])
        else:
            return None
        
    def get_folder_max_uid_next(self, id):
        if (self.box_list_map.has_key(id)):
            return int(self.box_list_map[id][6])
        else:
            return 1
        
    def addMessage(self, msg):
        pass
    
    def getMessage(self, id):
        pass

    def getVersion(self):
        pass
    
    def nextuid(self):
        self.lastuid += 1
        return self.lastuid
    
    def sync(self):
        self.connection.finish()

    def resync(self):
        self.sync()
        self.reopen()

    def getChanges(self, fromVersion):
        ids = self.getFolderIds()
        knownids = self.getGUIDs()
        for id in ids:
            folder_path = self.getFolderPath(id)
            folder_path_segments = folder_path.split('/')
            # determine parent folder
            parent_guid = None
            if (len(folder_path_segments) == 1):
                # root level folder
                parent_guid = ''
            else:
                parent_guid = self.getGUIDForFolderPath('/'.join(folder_path_segments[:-1]))
                if (parent_guid == None):
                    continue
            last_seen_folder_path = self.redis.hget(self.get_redis_basekey()+":guid2path", id)
            if (last_seen_folder_path != None):
                # check for changes
                if (folder_path != last_seen_folder_path):
                    # detected folder path change
                    last_seen_folder_path_segments = last_seen_folder_path.split('/')
                    if (folder_path_segments[-1] != last_seen_folder_path_segments[-1]):
                        # folder renamed
                        pass
                    if (folder_path_segments[:-1] != last_seen_folder_path_segments[:-1]):
                        # folder moved
                        pass
                    ics_key = id + ":" + "2" + ":" + self.getFolderLastChange(id)
                    self.redis.hset(self.get_redis_basekey()+":changes",ics_key,parent_guid + ':' + folder_path_segments[-1]) 
                    self.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key)
                try:
                    i = knownids.index(id)
                    del knownids[i]
                except:
                    pass
            else:
                source_key = self.redis.hget(self.get_redis_basekey()+":guid2sourceid", id)
                if (source_key != None):
                    # no previous path state set, create it below
                    pass
                else:
                    # unknown source_key, create folder
                    ics_key = id + ":" + "1" + ":" + self.getFolderLastChange(id)
                    self.redis.hset(self.get_redis_basekey()+":changes",ics_key,parent_guid + ':' + folder_path_segments[-1]) 
                    self.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key)
            # update last known path
            if (last_seen_folder_path != folder_path):
                self.redis.hset(self.get_redis_basekey()+":guid2path", id, folder_path)
        for knownid in knownids:
            tombstone = self.redis.hget(self.get_redis_basekey()+":tombstones",knownid)
            if (tombstone == None):
                # no tombstone, a freshly deleted folder
                ics_key = id + ":" + "3" + ":" + "0"
                self.redis.hset(self.get_redis_basekey()+":changes",ics_key,"0") 
                self.redis.rpush(self.get_redis_basekey()+":changeidqueue",ics_key)
                self.redis.hset(self.get_redis_basekey()+":tombstones",knownid,"0")
        pass
    
    def _getStoreStateData(self):
        connection = self.connection
        box_list_data = connection.boxlist()[0]
        box_list = []
        box_list_map = {}
        box_list_path_map = {}
        for box_list_item_line in box_list_data:
            box_list_item = box_list_item_line.split('\t', 9)
            box_list.append(box_list_item)
            if (len(box_list_item) > 4):
                id = box_list_item[4]
                box_list_map[id] = box_list_item
            path = box_list_item[0]
            box_list_path_map[path] = box_list_item
            if (box_list_item[0] == self.getReceiveFolderName()):
                self.ReceiveFolderId = box_list_item[4]
            elif (box_list_item[0] == self.getWastebasketFolderName()):
                self.WastebasketFolderId = box_list_item[4]
            elif (box_list_item[0] == self.getDraftsFolderName()):
                self.DraftsFolderId = box_list_item[4]
            elif (box_list_item[0] == self.getSentFolderName()):
                self.SentFolderId = box_list_item[4]
        self.box_list = box_list
        self.box_list_map = box_list_map
        self.box_list_path_map = box_list_path_map
        return box_list    

    def getReceiveFolderId(self):
        return self.ReceiveFolderId

    def getReceiveFolderName(self):
        return "INBOX"

    def getWastebasketFolderId(self):
        return self.WastebasketFolderId

    def getWastebasketFolderName(self):
        return "Trash"

    def getDraftsFolderId(self):
        return self.DraftsFolderId

    def getDraftsFolderName(self):
        return "Drafts"

    def getSentFolderId(self):
        return self.SentFolderId

    def getSentFolderName(self):
        return "Sent"

    def getAlternativeFolderId(self, guid):
        if (guid != self.WastebasketFolderId):
            return self.WastebasketFolderId
        else:
            return self.ReceiveFolderId
