# -*- coding: utf-8; indent-tabs-mode: nil -*-

# mailbridge: a mail storage spanning bridge
# 
# Copyright 2011 Ryan Hughes <ryan@n-h.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import traceback
import threading
import threadpool
import sys
import binascii

import MAPI
from MAPI.Util import *


def ReadUInt(stream):
    """Read one UInt from a binary stream.
    """
    return struct.unpack("I", stream.Read(4))[0]


class SyncState:
    """Represent a syncstate.
    
    A syncstate is defined as a four byte syncid and a four byte change id concatenated
    in a binary stream.
    """
    def __init__(self,arg):
        """Initialize a SyncState object
        
        arg: Can be an IStream object containing at least eight bytes or a struct
             containing the syncid and changeid.
        """
        if type(arg) == type(IStream()):
            clone = arg.Clone()
            clone.Seek(0, BOOKMARK_BEGINNING)
            self.syncid = ReadUInt(clone)
            self.changeid = ReadUInt(clone)
        else:
            self.syncid, self.changeid = struct.unpack("II", arg)

    def __getattr__(self, attr):
        if attr == 'syncid':
            return self.syncid
        elif attr == 'changeid':
            return self.changeid
        elif attr == 'key':
            return struct.pack("II", self.syncid, self.changeid)
        else:
            raise AttributeError

    def __repr__(self):
        return "SyncState: syncid=%u, changeid=%u" % (self.syncid, self.changeid)


class AdviseSink(ECChangeAdviseSink):
    def __init__(self, updater, collection):
        ECChangeAdviseSink.__init__(self, [IID_IECChangeAdviseSink])
        self.updater = updater
        self.collection = collection
        self.lock = threading.Lock()
        self.monitored = False
        self.events = {}
        self.jobinpool = 0

    def OnNotify(self, flags, entries):
        if len(entries) > 0:
            states = (SyncState(entry) for entry in entries)
            newevents = dict([state.syncid, state.changeid] for state in states)
            self.lock.acquire()
            exc = True
            try:
                try:
                    if self.monitored == True:
                        self.collection.AddEvents(self, newevents)
                    else:
                        self.events.update(newevents)
                except:
                    exc = False
                    if not self.lock.release():
                        raise
            finally:
                if exc:
                    self.lock.release()
        return 0
    
    def EnableMonitor(self):
        self.lock.acquire()
        exc = True
        try:
            try:
                if (self.monitored == False):
                    self.monitored = True
                    self.collection.AddEvents(self, self.events)
                    self.events = None
            except:
                exc = False
                if not self.lock.release():
                    raise
        finally:
            if exc:
                self.lock.release()

    def UpdateSyncState(self, state):
        self.collection.UpdateState(self, state)

    def SynchronizeFolder(self, syncid):
        self.updater.update_imap_cache_for_folder_by_syncid(syncid)


class AdviseSinkMonitor:
    def __init__(self):
        self.condition = threading.Condition()
        self.doExit = False
        self.sinkevents = {}
        self.notifsinkmonitor = None
        self.logger = logging.getLogger("AdviseSinkMonitor")

    def GetChangeAdviseSink(self, updater):
        return AdviseSink(updater, self)
    
    def getNotifSinkMonitor(self):
        if self.notifsinkmonitor != None:
            return self.notifsinkmonitor
        else:
            self.notifsinkmonitor = NotifSinkMonitor(self)
            return self.notifsinkmonitor

    def AddEvents(self, sink, newevents):
        self.logger.debug("Adding %u new events from sink %s" % (len(newevents), sink.__hash__()))
        self.condition.acquire()
        exc = True
        try:
            try:
                if (self.sinkevents.has_key(sink)):
                    events = self.sinkevents[sink]
                else:
                    self.sinkevents[sink] = events = {}
                events.update(newevents)
                self.condition.notify()
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def UpdateState(self, sink, state=None):
        self.logger.debug("Updating state for sink %s, syncid=%u, changeid=%u" % (sink.__hash__(), state.syncid, state.changeid))
        self.condition.acquire()
        exc = True
        try:
            try:
                events = self.sinkevents.get(sink, None)
                if events and events.has_key(state.syncid):
                    if events[state.syncid] <= state.changeid:
                        del events[state.syncid]
                        if not events:
                            del self.sinkevents[sink]
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def ProcessDone(self, sink):
        self.logger.debug("ProcessDone for sink %s" % (sink.__hash__()))
        self.condition.acquire()
        exc = True
        try:
            try:
                events = self.sinkevents.get(sink, None)
                sink.jobinpool = 0
                if events:
                    # there are newer events collected during update run
                    self.condition.notify()
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def Cancel(self):
        self.logger.debug("Canceling AdviseSinkMonitor...")
        self.condition.acquire()
        exc = True
        try:
            try:
                self.doExit = True
                self.condition.notify()
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()

    def Run(self, pool):
        self.logger.debug("AdviseSinkMonitor started")
        self.condition.acquire()
        exc = True
        try:
            try:
                while not self.doExit:
                    self.logger.debug("Waiting for events...")
                    self.condition.wait()
                    if not self.doExit:
                        self.logger.debug("Get events for %u sinks" % len(self.sinkevents))
                        requests = []
                        for item in self.sinkevents.items():
                            sink = item[0]
                            if not sink.jobinpool:
                                requests = requests + threadpool.makeRequests(self.ProcessSink, ([[item]] ))
                                sink.jobinpool = 1
                                del self.sinkevents[sink]
                        self.logger.debug("Put %d events" % len(requests))
                        [pool.putRequest(req) for req in requests]
            except:
                exc = False
                if not self.condition.release():
                    raise
        finally:
            if exc:
                self.condition.release()
    
    def ProcessSink(self, args):
        sink, events = args[0]
        self.logger.debug("Processing %u events for sink %s" % (len(events), sink.__hash__()))
        try:
            [sink.SynchronizeFolder(syncid) for syncid in events.keys()]
        except:
            self.logger.debug(traceback.format_exc())
        self.ProcessDone(sink)
                
            
class NotifSinkMonitor:
    def __init__(self, collection):
        self.collection = collection
        self.connections = []
        self.connectionstore = {}
        self.doExit = False
        self.notifsink = MAPINotifSink()
        self.updaters = {}
        self.sinkevents = {}
        self.logger = logging.getLogger("NotifSinkMonitor")

    def adviseStore(self, updater, store):
        store_guid = store.GetProps([PR_RECORD_KEY], 0)[0].Value
        self.updaters[store_guid] = updater
        connection = store.Advise(None, fnevObjectCreated|fnevObjectCopied|fnevObjectDeleted|fnevObjectModified|fnevObjectMoved, self.notifsink)
        self.connections.append(connection)
        self.connectionstore[connection] = store
        return self.notifsink

    def Cancel(self):
        self.logger.debug("Canceling NotifSinkMonitor...")
        self.doExit = True

    def Run(self, pool):
        self.logger.debug("NotifSinkMonitor started")
        exc = True
        try:
            try:
                while not self.doExit:
                    try:
                        self.logger.debug("Waiting for notifications...")
                        all_notifications = []
                        notifications = self.notifsink.GetNotifications(False, 5000)
                        while (not self.doExit and notifications):
                            all_notifications = all_notifications + notifications
                            notifications = self.notifsink.GetNotifications(False, 200)
                        if not self.doExit and all_notifications:
                            self.logger.debug("Collected %d notifications" % len(all_notifications))
                            seen_stores = {}
                            last_obj_type = {}
                            folder_changes = {}
                            for notification in all_notifications:
                                entryid = notification.lpEntryID
                                store_guid = entryid[4:20]
                                if notification.ulObjType == MAPI_FOLDER:
                                    if not folder_changes.has_key(store_guid):
                                        folder_changes[store_guid] = 1
                                        pool.putRequest(threadpool.WorkRequest(self.ProcessSink, [(store_guid)], None))
                                        self.logger.debug("put Request for store %s notification" % (binascii.hexlify(store_guid)))
                                seen_stores[store_guid] = 1
                    except:
                        traceback.print_exc(file=sys.stdout)
            except:
                exc = False
        finally:
            pass
                
        for connection in self.connections:
            self.connectionstore[connection].Unadvise(connection)
            del self.connectionstore[connection]
    
    def ProcessSink(self, args):
        store_guid = args
        self.logger.debug("Processing notification for store %s" % (binascii.hexlify(store_guid)))
        try:
            self.updaters[store_guid].check_update_imap_cache_hierarchy()
        except:
            traceback.print_exc(file=sys.stdout)
        